@@ -124,26 +124,38 @@ is_rwm (const char *str, libcrun_error_t *err)
 static int
 enable_controllers (oci_container_linux_resources *resources, const char *path, libcrun_error_t *err)
 {
+  cleanup_free char *controllers = NULL;
   cleanup_free char *tmp_path = NULL;
+  bool has_hugetlb = false;
+  bool has_memory = false;
+  bool has_cpuset = false;
+  bool has_pids = false;
+  bool has_cpu = false;
+  bool has_io = false;
   char *it;
   int ret;
-  cleanup_free char *controllers = NULL;
-  bool has_cpuset, has_cpu, has_io, has_memory, has_pids;
 
-  has_cpu = resources && resources->cpu && (resources->cpu->shares || resources->cpu->period
-                                            || resources->cpu->quota || resources->cpu->realtime_period
-                                            || resources->cpu->realtime_runtime);
-  has_cpuset = resources && resources->cpu && (resources->cpu->cpus || resources->cpu->mems);
-  has_io = resources && resources->block_io;
-  has_memory = resources && resources->memory;
-  has_pids = resources && resources->pids;
+  if (resources)
+    {
+      has_cpu = resources->cpu && (resources->cpu->shares
+                                   || resources->cpu->period
+                                   || resources->cpu->quota
+                                   || resources->cpu->realtime_period
+                                   || resources->cpu->realtime_runtime);
+      has_cpuset = resources->cpu && (resources->cpu->cpus || resources->cpu->mems);
+      has_io = resources->block_io;
+      has_memory = resources->memory;
+      has_pids = resources->pids;
+      has_hugetlb = resources->hugepage_limits_len;
+    }
 
-  xasprintf (&controllers, "%s %s %s %s %s",
+  xasprintf (&controllers, "%s %s %s %s %s %s",
              has_cpu ? "+cpu" : "",
              has_io ? "+io" : "",
              has_memory ? "+memory" : "",
              has_pids ? "+pids" : "",
-             has_cpuset ? "+cpuset" : "");
+             has_cpuset ? "+cpuset" : "",
+             has_hugetlb ? "+hugetlb" : "");
 
   xasprintf (&tmp_path, "%s/", path);
 
@@ -372,7 +384,7 @@ get_file_owner (const char *path, uid_t *uid, gid_t *gid)
   ret = statx (AT_FDCWD, path, AT_STATX_DONT_SYNC, STATX_UID|STATX_GID, &stx);
   if (UNLIKELY (ret < 0))
     {
-      if (errno == ENOSYS)
+      if (errno == ENOSYS || errno == EINVAL)
         goto fallback;
 
       return ret;
@@ -634,15 +646,20 @@ int systemd_finalize (oci_container_linux_resources *resources, int cgroup_mode,
       subpath = strchr (subsystem, ':') + 1;
       *(subpath - 1) = '\0';
 
-      /* skip named hierarchies that have no cgroup controller */
-      if (strcmp(subsystem, "") == 0)
-        continue;
-
       if (strcmp (subpath, *path))
         {
           ret = enter_cgroup_subsystem (pid, subsystem, *path, 1, err);
           if (UNLIKELY (ret < 0))
-            return ret;
+            {
+              /* If it is a named hierarchy, skip the error.  */
+              /* skip named hierarchies that have no cgroup controller */
+              if (strchr (subsystem, '='))
+                {
+                  crun_error_release (err);
+                  continue;
+                }
+              return ret;
+            }
         }
     }
 
@@ -1113,16 +1130,84 @@ libcrun_cgroup_pause_unpause (const char *cgroup_path, const bool pause, libcrun
   return libcrun_cgroup_pause_unpause_with_mode (cgroup_path, cgroup_mode, pause, err);
 }
 
-int
-libcrun_cgroup_read_pids (const char *path, pid_t **pids, libcrun_error_t *err)
+static
+int read_pids_cgroup (int dfd, bool recurse, pid_t **pids, size_t *n_pids, size_t *allocated, libcrun_error_t *err)
 {
-  cleanup_free char *cgroup_path_procs = NULL;
+  cleanup_close int clean_dfd = dfd;
+  cleanup_close int tasksfd = -1;
   cleanup_free char *buffer = NULL;
-  int ret;
-  size_t len;
-  size_t n_pids;
-  char *it;
   char *saveptr = NULL;
+  size_t n_new_pids;
+  size_t len;
+  char *it;
+  int ret;
+
+  tasksfd = openat (dfd, "cgroup.procs", O_RDONLY|O_CLOEXEC);
+  if (tasksfd < 0)
+    return crun_make_error (err, errno, "open cgroup.procs");
+
+  ret = read_all_fd (tasksfd, "cgroup.procs", &buffer, &len, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  for (n_new_pids = 0, it = buffer; it; it = strchr (it + 1, '\n'))
+    n_new_pids++;
+
+  if (*allocated < *n_pids + n_new_pids + 1)
+    {
+      *allocated = *n_pids + n_new_pids + 1;
+      *pids = xrealloc (*pids, sizeof (pid_t) * *allocated);
+    }
+
+  for (it = strtok_r (buffer, "\n", &saveptr); it; it = strtok_r (NULL, "\n", &saveptr))
+    {
+      pid_t pid = strtoul (it, NULL, 10);
+
+      if (pid > 0)
+        (*pids)[(*n_pids)++] = pid;
+    }
+  (*pids)[*n_pids] = 0;
+
+  if (recurse)
+    {
+      cleanup_dir DIR *dir = NULL;
+      struct dirent *de;
+
+      dir = fdopendir (dfd);
+      if (UNLIKELY (dir == NULL))
+          return crun_make_error (err, errno, "open cgroup sub-directory");
+      /* Now dir owns the dfd descriptor.  */
+      clean_dfd = -1;
+
+      for (de = readdir (dir); de; de = readdir (dir))
+        {
+          int nfd;
+
+          if (strcmp (de->d_name, ".") == 0 ||
+              strcmp (de->d_name, "..") == 0)
+            continue;
+
+          if (de->d_type != DT_DIR)
+            continue;
+
+          nfd = openat (dirfd (dir), de->d_name, O_DIRECTORY|O_CLOEXEC);
+          if (UNLIKELY (dir == NULL))
+            return crun_make_error (err, errno, "open cgroup directory %s", de->d_name);
+          ret = read_pids_cgroup (nfd, recurse, pids, n_pids, allocated, err);
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
+    }
+  return 0;
+}
+
+int
+libcrun_cgroup_read_pids (const char *path, bool recurse, pid_t **pids, libcrun_error_t *err)
+{
+  cleanup_free char *cgroup_path = NULL;
+  cleanup_free char *buffer = NULL;
+  size_t n_pids, allocated;
+  int dirfd;
   int mode;
 
   if (path == NULL || *path == '\0')
@@ -1135,35 +1220,26 @@ libcrun_cgroup_read_pids (const char *path, pid_t **pids, libcrun_error_t *err)
   switch (mode)
     {
     case CGROUP_MODE_UNIFIED:
-      xasprintf (&cgroup_path_procs, "/sys/fs/cgroup/%s/cgroup.procs", path);
-      ret = read_all_file (cgroup_path_procs, &buffer, &len, err);
-      if (UNLIKELY (ret < 0))
-        return ret;
+      xasprintf (&cgroup_path, "/sys/fs/cgroup/%s", path);
       break;
 
     case CGROUP_MODE_HYBRID:
     case CGROUP_MODE_LEGACY:
-      xasprintf (&cgroup_path_procs, "/sys/fs/cgroup/pids/%s/cgroup.procs", path);
-      ret = read_all_file (cgroup_path_procs, &buffer, &len, err);
-      if (UNLIKELY (ret < 0))
-        return ret;
+      xasprintf (&cgroup_path, "/sys/fs/cgroup/pids/%s", path);
       break;
+
+    default:
+      return crun_make_error (err, errno, "invalid cgroup mode %d", mode);
     }
 
-  for (n_pids = 0, it = buffer; it; it = strchr (it + 1, '\n'))
-    n_pids++;
+  dirfd = open (cgroup_path, O_DIRECTORY | O_CLOEXEC);
+  if (dirfd < 0)
+    return crun_make_error (err, errno, "open %s", cgroup_path);
 
-  *pids = xmalloc (sizeof (pid_t) * (n_pids + 1));
+  n_pids = 0;
+  allocated = 0;
 
-  for (n_pids = 0, it = strtok_r (buffer, "\n", &saveptr); it; it = strtok_r (NULL, "\n", &saveptr))
-    {
-      pid_t pid = strtoul (it, NULL, 10);
-      if (pid > 0)
-        (*pids)[n_pids++] = pid;
-    }
-  (*pids)[n_pids] = 0;
-
-  return 0;
+  return read_pids_cgroup (dirfd, recurse, pids, &n_pids, &allocated, err);
 }
 
 int
@@ -1180,11 +1256,17 @@ libcrun_cgroup_killall_signal (char *path, int signal, libcrun_error_t *err)
   if (UNLIKELY (ret < 0))
     crun_error_release (err);
 
-  ret = libcrun_cgroup_read_pids (path, &pids, err);
+  ret = libcrun_cgroup_read_pids (path, true, &pids, err);
   if (UNLIKELY (ret < 0))
-    return ret;
+    {
+      if (crun_error_get_errno (err) != ENOENT)
+        return ret;
 
-  for (i = 0; pids[i]; i++)
+      /* If the file doesn't exist then the container was already killed.  */
+      crun_error_release (err);
+    }
+
+  for (i = 0; pids && pids[i]; i++)
     {
       ret = kill (pids[i], signal);
       if (UNLIKELY (ret < 0))
@@ -1540,17 +1622,20 @@ write_network_resources (int dirfd, oci_container_linux_resources_network *net,
 }
 
 static int
-write_hugetlb_resources (int dirfd, oci_container_linux_resources_hugepage_limits_element **htlb, size_t htlb_len, libcrun_error_t *err)
+write_hugetlb_resources (int dirfd, bool cgroup2, oci_container_linux_resources_hugepage_limits_element **htlb, size_t htlb_len, libcrun_error_t *err)
 {
   char fmt_buf[128];
   size_t i;
   for (i = 0; i < htlb_len; i++)
     {
       cleanup_free char *filename = NULL;
+      const char *suffix;
       size_t len;
       int ret;
 
-      xasprintf (&filename, "hugetlb.%s.limit_in_bytes", htlb[i]->page_size);
+      suffix = cgroup2 ? "max" : "limit_in_bytes";
+
+      xasprintf (&filename, "hugetlb.%s.%s", htlb[i]->page_size, suffix);
 
       len = sprintf (fmt_buf, "%lu", htlb[i]->limit);
       ret = write_file_at (dirfd, filename, fmt_buf, len, err);
@@ -1783,7 +1868,7 @@ write_memory_resources (int dirfd, bool cgroup2, oci_container_linux_resources_m
   if (memory->reservation)
     {
       len = sprintf (fmt_buf, "%lu", memory->reservation);
-      ret = write_file_at (dirfd, cgroup2 ? "memory.high" : "memory.soft_limit_in_bytes", fmt_buf, len, err);
+      ret = write_file_at (dirfd, cgroup2 ? "memory.low" : "memory.soft_limit_in_bytes", fmt_buf, len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -1988,7 +2073,7 @@ update_cgroup_v1_resources (oci_container_linux_resources *resources, char *path
       if (UNLIKELY (dirfd_htlb < 0))
         return crun_make_error (err, errno, "open %s", path_to_htlb);
 
-      ret = write_hugetlb_resources (dirfd_htlb,
+      ret = write_hugetlb_resources (dirfd_htlb, false,
                                      resources->hugepage_limits,
                                      resources->hugepage_limits_len,
                                      err);
@@ -2091,8 +2176,6 @@ update_cgroup_v2_resources (oci_container_linux_resources *resources, char *path
 
   if (resources->network)
     return crun_make_error (err, 0, "network limits not supported on cgroupv2");
-  if (resources->hugepage_limits_len)
-    return crun_make_error (err, 0, "hugepages not supported on cgroupv2");
 
   xasprintf (&cgroup_path, "/sys/fs/cgroup%s", path);
 
@@ -2147,6 +2230,16 @@ update_cgroup_v2_resources (oci_container_linux_resources *resources, char *path
         return ret;
     }
 
+  if (resources->hugepage_limits_len)
+    {
+      ret = write_hugetlb_resources (cgroup_dirfd, true,
+                                     resources->hugepage_limits,
+                                     resources->hugepage_limits_len,
+                                     err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+
   return 0;
 }
 
