@@ -64,7 +64,10 @@ struct container_entrypoint_s
   libcrun_context_t *context;
   int has_terminal_socket_pair;
   int terminal_socketpair[2];
+
+  /* Used by log_write_to_sync_socket.  */
   int sync_socket;
+
   int seccomp_fd;
   int console_socket_fd;
 };
@@ -463,11 +466,12 @@ initialize_security (oci_container_process *proc, libcrun_error_t *err)
   return 0;
 }
 
-
+/* Initialize the environment where the container process runs.
+   It is used by the container init process.  */
 static int
-container_entrypoint_init (void *args, const char *notify_socket,
-                           int sync_socket, const char **exec_path,
-                           libcrun_error_t *err)
+container_init_setup (void *args, const char *notify_socket,
+                      int sync_socket, const char **exec_path,
+                      libcrun_error_t *err)
 {
   struct container_entrypoint_s *entrypoint_args = args;
   libcrun_container_t *container = entrypoint_args->container;
@@ -606,6 +610,13 @@ container_entrypoint_init (void *args, const char *notify_socket,
   if (UNLIKELY (ret < 0))
     return ret;
 
+  if (container->container_def->linux && container->container_def->linux->personality)
+    {
+      ret = libcrun_set_personality (container->container_def->linux->personality, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+
   if (def->process && !def->process->no_new_privileges)
     {
       char **seccomp_flags = NULL;
@@ -617,7 +628,7 @@ container_entrypoint_init (void *args, const char *notify_socket,
           seccomp_flags_len = def->linux->seccomp->flags_len;
         }
 
-      ret = libcrun_generate_and_load_seccomp (entrypoint_args->container, entrypoint_args->seccomp_fd, seccomp_flags, seccomp_flags_len, err);
+      ret = libcrun_apply_seccomp (entrypoint_args->seccomp_fd, seccomp_flags, seccomp_flags_len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -641,8 +652,8 @@ container_entrypoint_init (void *args, const char *notify_socket,
 
 /* Entrypoint to the container.  */
 static int
-container_entrypoint (void *args, const char *notify_socket,
-                      int sync_socket, libcrun_error_t *err)
+container_init (void *args, const char *notify_socket, int sync_socket,
+                libcrun_error_t *err)
 {
   struct container_entrypoint_s *entrypoint_args = args;
   int ret;
@@ -652,7 +663,7 @@ container_entrypoint (void *args, const char *notify_socket,
 
   crun_set_output_handler (log_write_to_sync_socket, args, false);
 
-  ret = container_entrypoint_init (args, notify_socket, sync_socket, &exec_path, err);
+  ret = container_init_setup (args, notify_socket, sync_socket, &exec_path, err);
   if (UNLIKELY (ret < 0))
     {
       /* If it fails to write the error using the sync socket, then fallback
@@ -713,7 +724,7 @@ container_entrypoint (void *args, const char *notify_socket,
           seccomp_flags_len = def->linux->seccomp->flags_len;
         }
 
-      ret = libcrun_generate_and_load_seccomp (entrypoint_args->container, entrypoint_args->seccomp_fd, seccomp_flags, seccomp_flags_len, err);
+      ret = libcrun_apply_seccomp (entrypoint_args->seccomp_fd, seccomp_flags, seccomp_flags_len, err);
       if (UNLIKELY (ret < 0))
         return ret;
     }
@@ -724,9 +735,15 @@ container_entrypoint (void *args, const char *notify_socket,
   if (UNLIKELY (exec_path == NULL))
     return crun_make_error (err, errno, "executable path not specified");
 
+  if (def->process->user)
+    umask (def->process->user->umask);
+
   execv (exec_path, def->process->args);
 
-  return crun_make_error (err, errno, "exec the container process");
+  if (errno == ENOENT)
+    return crun_make_error (err, errno, "exec container process (missing dynamic library?) `%s`", exec_path);
+
+  return crun_make_error (err, errno, "exec container process `%s`", exec_path);
 }
 
 static int
@@ -1404,7 +1421,7 @@ libcrun_container_run_internal (libcrun_container_t *container, libcrun_context_
     }
 
   pid = libcrun_run_linux_container (container, context->detach,
-                                     container_entrypoint, &container_args,
+                                     container_init, &container_args,
                                      &sync_socket, err);
   if (UNLIKELY (pid < 0))
     return pid;
@@ -1417,9 +1434,6 @@ libcrun_container_run_internal (libcrun_container_t *container, libcrun_context_
         return ret;
     }
 
-  if (seccomp_fd >= 0)
-    close_and_reset (&seccomp_fd);
-
   if (container_args.terminal_socketpair[1] >= 0)
     close_and_reset (&socket_pair_1);
 
@@ -1455,6 +1469,24 @@ libcrun_container_run_internal (libcrun_container_t *container, libcrun_context_
         }
     }
 
+  if (seccomp_fd >= 0)
+    {
+       unsigned int seccomp_gen_options = 0;
+       const char *annotation;
+
+       annotation = find_annotation (container, "run.oci.seccomp_fail_unknown_syscall");
+       if (annotation && strcmp (annotation, "0") != 0)
+         seccomp_gen_options = LIBCRUN_SECCOMP_FAIL_UNKNOWN_SYSCALL;
+
+       ret = libcrun_generate_seccomp (container, seccomp_fd, seccomp_gen_options, err);
+      if (UNLIKELY (ret < 0))
+        {
+          cleanup_watch (context, pid, def, context->id, sync_socket, terminal_fd);
+          return ret;
+        }
+      close_and_reset (&seccomp_fd);
+    }
+
   ret = sync_socket_send_sync (sync_socket, true, err);
   if (UNLIKELY (ret < 0))
     {
@@ -1687,7 +1719,7 @@ libcrun_container_run (libcrun_context_t *context, libcrun_container_t *containe
 }
 
 int
-libcrun_container_create (libcrun_context_t *context, libcrun_container_t *container, libcrun_error_t *err)
+libcrun_container_create (libcrun_context_t *context, libcrun_container_t *container, unsigned int options, libcrun_error_t *err)
 {
   oci_container *def = container->container_def;
   int ret;
@@ -1717,6 +1749,17 @@ libcrun_container_create (libcrun_context_t *context, libcrun_container_t *conta
   if (UNLIKELY (exec_fifo_fd < 0))
     return exec_fifo_fd;
 
+  context->fifo_exec_wait_fd = exec_fifo_fd;
+  exec_fifo_fd = -1;
+
+  if ((options & LIBCRUN_RUN_OPTIONS_PREFORK) == 0)
+    {
+      ret = libcrun_copy_config_file (context->id, context->state_root, context->bundle, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+      return libcrun_container_run_internal (container, context, -1, err);
+    }
+
   ret = pipe (container_ready_pipe);
   if (UNLIKELY (ret < 0))
     return crun_make_error (err, errno, "pipe");
@@ -1749,9 +1792,6 @@ libcrun_container_create (libcrun_context_t *context, libcrun_container_t *conta
       return 1;
     }
 
-  context->fifo_exec_wait_fd = exec_fifo_fd;
-  exec_fifo_fd = -1;
-
   /* forked process.  */
   ret = detach_process ();
   if (UNLIKELY (ret < 0))
@@ -2077,6 +2117,17 @@ libcrun_container_exec (libcrun_context_t *context, const char *id, oci_containe
             libcrun_warning ("cannot set HOME environment variable");
         }
 
+      /* If the new process block doesn't specify a SELinux label or AppArmor profile, then
+         use the configuration from the original config file.  */
+      if (container->container_def->process)
+        {
+          if (process->selinux_label == NULL && container->container_def->process->selinux_label)
+            process->selinux_label = container->container_def->process->selinux_label;
+
+          if (process->apparmor_profile == NULL && container->container_def->process->apparmor_profile)
+            process->apparmor_profile = container->container_def->process->apparmor_profile;
+        }
+
       if (UNLIKELY (libcrun_set_selinux_exec_label (process, err) < 0))
         libcrun_fail_with_error ((*err)->status, "%s", (*err)->msg);
 
@@ -2100,6 +2151,13 @@ libcrun_container_exec (libcrun_context_t *context, const char *id, oci_containe
           libcrun_fail_with_error ((*err)->status, "%s", (*err)->msg);
         }
 
+      if (container->container_def->linux && container->container_def->linux->personality)
+        {
+          ret = libcrun_set_personality (container->container_def->linux->personality, err);
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
+
       if (!process->no_new_privileges)
         {
           ret = libcrun_apply_seccomp (seccomp_fd, seccomp_flags, seccomp_flags_len, err);
@@ -2139,6 +2197,9 @@ libcrun_container_exec (libcrun_context_t *context, const char *id, oci_containe
             return ret;
         }
 
+      if (process->user)
+        umask (process->user->umask);
+
       TEMP_FAILURE_RETRY (write (pipefd1, "0", 1));
       close (pipefd1);
       pipefd1 = -1;
