@@ -44,6 +44,8 @@
 #include <sys/wait.h>
 #include <sys/vfs.h>
 #include <limits.h>
+#include <inttypes.h>
+#include <sys/personality.h>
 
 #ifndef RLIMIT_RTTIME
 # define RLIMIT_RTTIME 15
@@ -1377,23 +1379,6 @@ deny_setgroups (libcrun_container_t *container, pid_t pid, libcrun_error_t *err)
   return ret;
 }
 
-static const char *
-find_annotation (libcrun_container_t *container, const char *name)
-{
-  size_t i;
-
-  if (container->container_def->annotations == NULL)
-    return NULL;
-
-  for (i = 0; i < container->container_def->annotations->len; i++)
-    {
-      if (strcmp (container->container_def->annotations->keys[i], name) == 0)
-        return container->container_def->annotations->values[i];
-    }
-
-  return NULL;
-}
-
 static int
 can_setgroups (libcrun_container_t *container, libcrun_error_t *err)
 {
@@ -1408,7 +1393,7 @@ can_setgroups (libcrun_container_t *container, libcrun_error_t *err)
       const char *annotation;
 
       /* Skip setgroups if the annotation is set to anything different than "0".  */
-      annotation = find_annotation (container, "io.crun.keep_original_groups");
+      annotation = find_annotation (container, "run.oci.keep_original_groups");
       if (annotation)
         return strcmp (annotation, "0") == 0 ? 1 : 0;
     }
@@ -1440,6 +1425,8 @@ libcrun_container_enter_cgroup_ns (libcrun_container_t *container, libcrun_error
 int
 libcrun_set_usernamespace (libcrun_container_t *container, pid_t pid, libcrun_error_t *err)
 {
+#define MAPPING_FMT_SIZE ("%" PRIu32 " %" PRIu32 " %" PRIu32 "\n")
+#define MAPPING_FMT_1 ("%" PRIu32 " %" PRIu32 " 1\n")
   cleanup_free char *uid_map_file = NULL;
   cleanup_free char *gid_map_file = NULL;
   cleanup_free char *uid_map = NULL;
@@ -1457,9 +1444,9 @@ libcrun_set_usernamespace (libcrun_container_t *container, pid_t pid, libcrun_er
       if (uid_map == NULL)
         {
           if (container->host_uid)
-            uid_map_len = xasprintf (&uid_map, "%d %d 1", 0, container->host_uid);
+            uid_map_len = xasprintf (&uid_map, MAPPING_FMT_1, 0, container->host_uid);
           else
-            uid_map_len = xasprintf (&uid_map, "%d %d %d", 0, container->host_uid, container->container_uid + 1);
+            uid_map_len = xasprintf (&uid_map, MAPPING_FMT_SIZE, 0, container->host_uid, container->container_uid + 1);
         }
     }
   else
@@ -1471,7 +1458,7 @@ libcrun_set_usernamespace (libcrun_container_t *container, pid_t pid, libcrun_er
         {
           size_t len;
 
-          len = sprintf (buffer, "%d %d %d\n",
+          len = sprintf (buffer, MAPPING_FMT_SIZE,
                          def->linux->uid_mappings[s]->container_id,
                          def->linux->uid_mappings[s]->host_id,
                          def->linux->uid_mappings[s]->size);
@@ -1488,9 +1475,9 @@ libcrun_set_usernamespace (libcrun_container_t *container, pid_t pid, libcrun_er
       if (gid_map == NULL)
         {
           if (container->host_gid)
-            gid_map_len = xasprintf (&gid_map, "%d %d 1", container->container_gid, container->host_gid);
+            gid_map_len = xasprintf (&gid_map, MAPPING_FMT_1, container->container_gid, container->host_gid);
           else
-            gid_map_len = xasprintf (&gid_map, "%d %d %d", 0, container->host_gid, container->container_gid + 1);
+            gid_map_len = xasprintf (&gid_map, MAPPING_FMT_SIZE, 0, container->host_gid, container->container_gid + 1);
         }
     }
   else
@@ -1502,7 +1489,7 @@ libcrun_set_usernamespace (libcrun_container_t *container, pid_t pid, libcrun_er
         {
           size_t len;
 
-          len = sprintf (buffer, "%d %d %d\n",
+          len = sprintf (buffer, MAPPING_FMT_SIZE,
                          def->linux->gid_mappings[s]->container_id,
                          def->linux->gid_mappings[s]->host_id,
                          def->linux->gid_mappings[s]->size);
@@ -1531,7 +1518,7 @@ libcrun_set_usernamespace (libcrun_container_t *container, pid_t pid, libcrun_er
           if (UNLIKELY (ret < 0))
             return ret;
 
-          single_mapping_len = sprintf (single_mapping, "%d %d 1", container->container_gid, container->host_gid);
+          single_mapping_len = sprintf (single_mapping, MAPPING_FMT_1, container->container_gid, container->host_gid);
           ret = write_file (gid_map_file, single_mapping, single_mapping_len, err);
         }
     }
@@ -1559,13 +1546,16 @@ libcrun_set_usernamespace (libcrun_container_t *container, pid_t pid, libcrun_er
                 return ret;
             }
 
-          single_mapping_len = sprintf (single_mapping, "%d %d 1", container->container_uid, container->host_uid);
+          single_mapping_len = sprintf (single_mapping, MAPPING_FMT_1, container->container_uid, container->host_uid);
           ret = write_file (uid_map_file, single_mapping, single_mapping_len, err);
         }
     }
   if (UNLIKELY (ret < 0))
     return ret;
   return 0;
+
+#undef MAPPING_FMT_SIZE
+#undef MAPPING_FMT_1
 }
 
 #define CAP_TO_MASK_0(x) (1L << ((x) & 31))
@@ -2540,6 +2530,26 @@ libcrun_container_unpause_linux (libcrun_container_status_t *status, libcrun_err
   return libcrun_container_pause_unpause_linux (status, false, err);
 }
 
+int
+libcrun_set_personality (oci_container_linux_personality *p, libcrun_error_t *err)
+{
+  unsigned long persona = 0;
+  int ret;
+
+  if (strcmp (p->domain, "LINUX") == 0)
+      persona = PER_LINUX;
+  else if (strcmp (p->domain, "LINUX32") == 0)
+      persona = PER_LINUX32;
+  else
+    return crun_make_error (err, 0, "unknown persona specified '%s'", p->domain);
+
+  ret = personality (persona);
+  if (UNLIKELY (ret < 0))
+    return crun_make_error (err, 0, "set personality to '%s'", p->domain);
+
+  return 0;
+}
+
 /* Protection for attacks like CVE-2019-5736.  */
 int ensure_cloned_binary ();
 __attribute__((constructor)) static void libcrun_rexec(void)
