@@ -207,12 +207,55 @@ detach_process ()
   return 0;
 }
 
+static int
+get_file_type_at (int dirfd, mode_t *mode, bool nofollow, const char *path)
+{
+  struct stat st;
+  int ret;
+
+#ifdef HAVE_STATX
+  struct statx stx;
+
+  ret = statx (dirfd, path, (nofollow ? AT_SYMLINK_NOFOLLOW : 0) | AT_STATX_DONT_SYNC, STATX_TYPE, &stx);
+  if (UNLIKELY (ret < 0))
+    {
+      if (errno == ENOSYS || errno == EINVAL)
+        goto fallback;
+
+      return ret;
+    }
+  *mode = stx.stx_mode;
+  return ret;
+
+ fallback:
+#endif
+  ret = fstatat (dirfd, path, &st, nofollow ? AT_SYMLINK_NOFOLLOW : 0);
+  *mode = st.st_mode;
+  return ret;
+}
+
+static int
+get_file_type (mode_t *mode, bool nofollow, const char *path)
+{
+  return get_file_type_at (AT_FDCWD, mode, nofollow, path);
+}
+
 int
 create_file_if_missing_at (int dirfd, const char *file, libcrun_error_t *err)
 {
   cleanup_close int fd_write = openat (dirfd, file, O_CREAT | O_WRONLY, 0700);
   if (fd_write < 0)
-    return crun_make_error (err, errno, "creating file '%s'", file);
+    {
+      mode_t mode;
+      int ret;
+
+      /* On errors, check if the file already exists.  */
+      ret = get_file_type_at (dirfd, &mode, false, file);
+      if (ret == 0 && S_ISREG (mode))
+        return 0;
+
+      return crun_make_error (err, errno, "creating file '%s'", file);
+    }
   return 0;
 }
 
@@ -221,7 +264,17 @@ create_file_if_missing (const char *file, libcrun_error_t *err)
 {
   cleanup_close int fd_write = open (file, O_CREAT | O_WRONLY, 0700);
   if (fd_write < 0)
-    return crun_make_error (err, errno, "creating file '%s'", file);
+    {
+      mode_t mode;
+      int ret;
+
+      /* On errors, check if the file already exists.  */
+      ret = get_file_type (&mode, false, file);
+      if (ret == 0 && S_ISREG (mode))
+        return 0;
+
+      return crun_make_error (err, errno, "creating file '%s'", file);
+    }
   return 0;
 }
 
@@ -245,7 +298,14 @@ ensure_directory_internal (char *path, size_t len, int mode, libcrun_error_t *er
         }
 
       if (errno != ENOENT || parent_created)
-        return crun_make_error (err, errno, "creating file '%s'", path);
+        {
+          /* On errors check if the directory already exists.  */
+          ret = crun_dir_p (path, false, err);
+          if (ret > 0)
+            break;
+
+          return crun_make_error (err, errno, "creating file '%s'", path);
+        }
       else
         {
           while (it > path && *it != '/')
@@ -322,7 +382,7 @@ get_file_size (int fd, off_t *size)
   ret = statx (fd, "", AT_EMPTY_PATH | AT_SYMLINK_NOFOLLOW | AT_STATX_DONT_SYNC, STATX_SIZE, &stx);
   if (UNLIKELY (ret < 0))
     {
-      if (errno == ENOSYS)
+      if (errno == ENOSYS || errno == EINVAL)
         goto fallback;
       return ret;
     }
@@ -337,38 +397,6 @@ get_file_size (int fd, off_t *size)
   return ret;
 }
 
-static int
-get_file_type (mode_t *mode, bool nofollow, const char *path)
-{
-  struct stat st;
-  int ret;
-
-#ifdef HAVE_STATX
-  struct statx stx;
-
-  ret = statx (AT_FDCWD, path, (nofollow ? AT_SYMLINK_NOFOLLOW : 0) | AT_STATX_DONT_SYNC, STATX_TYPE, &stx);
-  if (UNLIKELY (ret < 0))
-    {
-      if (errno == ENOSYS)
-        goto fallback;
-
-      return ret;
-    }
-  *mode = stx.stx_mode;
-  return ret;
-
- fallback:
-#endif
-  if (nofollow)
-    ret = lstat (path, &st);
-  else
-    ret = stat (path, &st);
-  if (UNLIKELY (ret < 0))
-    return ret;
-  *mode = st.st_mode;
-  return ret;
-}
-
 int
 crun_dir_p (const char *path, bool nofollow, libcrun_error_t *err)
 {
@@ -1091,15 +1119,27 @@ run_process_with_stdin_timeout_envp (char *path,
       r = TEMP_FAILURE_RETRY (waitpid (pid, &status, 0));
       if (r < 0)
         return crun_make_error (err, errno, "waitpid");
-      if (WIFEXITED (status) || WIFSIGNALED (status))
+      if (WIFEXITED (status))
         return WEXITSTATUS (status);
+      if (WIFSIGNALED (status))
+        return 127 + WTERMSIG (status);
     }
   else
     {
       char *tmp_args[] = {path, NULL};
+      int out = open ("/dev/null", O_WRONLY);
+
+      if (UNLIKELY (out < 0))
+        _exit (EXIT_FAILURE);
+
       close (pipe_w);
       dup2 (pipe_r, 0);
       close (pipe_r);
+
+      dup2 (out, 1);
+      dup2 (out, 2);
+      close (out);
+
       if (args == NULL)
         args = tmp_args;
 
@@ -1400,7 +1440,7 @@ int copy_rec_stat_file_at (int dfd, const char *path, mode_t *mode, off_t *size,
   ret = statx (dfd, path, AT_SYMLINK_NOFOLLOW | AT_STATX_DONT_SYNC, STATX_TYPE | STATX_MODE | STATX_SIZE | STATX_UID | STATX_GID, &stx);
   if (UNLIKELY (ret < 0))
     {
-      if (errno == ENOSYS)
+      if (errno == ENOSYS || errno == EINVAL)
         goto fallback;
 
       return ret;
@@ -1578,3 +1618,20 @@ copy_recursive_fd_to_fd (int srcdirfd, int destdirfd, const char *srcname, const
 
   return 0;
 }
+
+const char *
+find_annotation (libcrun_container_t *container, const char *name)
+{
+  size_t i;
+
+  if (container->container_def->annotations == NULL)
+    return NULL;
+
+  for (i = 0; i < container->container_def->annotations->len; i++)
+    {
+      if (strcmp (container->container_def->annotations->keys[i], name) == 0)
+        return container->container_def->annotations->values[i];
+    }
+
+  return NULL;
+}
